@@ -3,6 +3,7 @@ from __future__ import absolute_import
 import time
 import string
 import logging
+from collections import namedtuple
 from contextlib import contextmanager
 
 import xapian
@@ -21,6 +22,11 @@ RETRY_LIMIT = 3
 
 logger = logging.getLogger(__name__)
 
+XAPIAN_VERSION = namedtuple('XAPIAN_VERSION', ['major', 'minor', 'revision'])
+
+xapian_version = XAPIAN_VERSION(xapian.major_version(),
+                                  xapian.minor_version(),
+                                  xapian.revision())
 
 default_parser_flags = (QueryParser.FLAG_PHRASE |
                         QueryParser.FLAG_BOOLEAN |
@@ -52,8 +58,12 @@ def _prefix(name):
 class Record(object):
     """Nice attribute-accessable record for a search result."""
 
-    def __init__(self, schema):
+    def __init__(self, schema, document, percent, rank, weight):
         self._xodb_schema = schema
+        self._xodb_document = document
+        self._xodb_percent = percent
+        self._xodb_rank = rank
+        self._xodb_weight = weight
 
     def __getattr__(self, name):
         try:
@@ -65,8 +75,8 @@ class Record(object):
         return repr(self._xodb_schema)
 
 
-def record_factory(database, record):
-    return Record(record)
+def record_factory(database, data, doc, percent, rank, weight):
+    return Record(data, doc, percent, rank, weight)
 
 
 class LanguageDecider(xapian.ExpandDecider):
@@ -126,23 +136,29 @@ class MultipleValueRangeProcessor(xapian.ValueRangeProcessor):
 
     :param map: a dict of prefix to value number pairs.
 
-    :param serializer: optional callable to serialize the arguments into
+    :param serializer: optional callable to serialize the range arguments into
                        the same form as the corresponding values are stored.
                        Typically xapian.sortable_serialise for floats,
                        str.lower for strings.
+
+    :param end_serializer: optional callable to serialize the end argument
+                           of the range into the same form as the values.
+                           Defaults to the value of *serializer*.
+
     """
 
-    def __init__(self, map, serializer=None):
+    def __init__(self, map, serializer=None, end_serializer=None):
         self.map = map
-        self.serializer = serializer or (lambda x: x)
+        self.begin_serializer = serializer or (lambda x: x)
+        self.end_serializer = end_serializer or self.begin_serializer
         xapian.ValueRangeProcessor.__init__(self)
 
     def __call__(self, begin, end):
         for prefix, value in self.map.items():
             if begin.startswith(prefix + ':'):
                 return (value,
-                        self.serializer(begin[len(prefix) + 1:]),
-                        self.serializer(end))
+                        self.begin_serializer(begin[len(prefix) + 1:]),
+                        self.end_serializer(end))
         return (xapian.BAD_VALUENO, begin, end)
 
 
@@ -286,19 +302,28 @@ class Database(object):
         self.boolean_prefixes[key] = value
         self.backend.set_metadata(self.boolean_prefix + key, value)
 
+    def allocate_value_index(self, name):
+        """Default implementation of value index number allocation.
+
+        Implement your own using shared something to keep several
+        databases in sync. i.e. for indexing in parallel.
+        """
+        value_count = self.value_count + 1
+        self.value_count = value_count
+        return value_count
+
     def add_value(self, name, sort=None):
         """Add a value mapping to the database.
         """
         if name in self.values:
             return self.values[name]
-        value_count = self.value_count + 1
-        self.value_count = value_count
-        self.values[name] = value_count
-        self.backend.set_metadata(self.value_prefix + name, str(value_count))
+        value_index = self.allocate_value_index(name)
+        self.values[name] = value_index
+        self.backend.set_metadata(self.value_prefix + name, str(value_index))
         if sort:
             self.value_sorts[name] = sort
             self.backend.set_metadata(self.value_sort_prefix + name, sort)
-        return value_count
+        return value_index
 
     def __len__(self):
         """ Return the number of documents in this database. """
@@ -363,33 +388,68 @@ class Database(object):
 
         :params objs: One or more mapped objects to add.
 
+        :param schema_type: Specify the schema to be used. (optional)
+
         :param validate: Validated the schema before the object is
         added.  Default: True
 
         Returns a list of xapan documents that were added to the
         database.
         """
+        assert self._writable, "Database is not writable"
         added = []
         validate = kw.pop('validate', True)
+        schema_type = kw.pop('schema_type', None)
         for obj in objs:
             if isinstance(obj, xapian.Document):
                 self.backend.add_document(obj)
                 added.append(obj)
                 continue
-            doc = self.to_document(obj, validate=validate)
+            doc = self.to_document(obj, schema_type=schema_type,
+                                   validate=validate)
             self.backend.add_document(doc)
             added.append(doc)
         return added
 
-    def to_schema(self, obj, validate=True):
+    def replace(self, obj, docid, **kw):
+        """Add or replace an object in the database with a specified
+        document id by transforming it into a xapian document.
+        It's type or one of its base types must be
+        mapped to a schema before an object can be added here.
+
+        :param obj: The mapped objects to add.
+
+        :param docid:  The document id to user for the mapped object.
+
+        :param schema_type: Specify the schema to be used. (optional)
+
+        :param validate: Validated the schema before the object is
+        added.  Default: True
+
+        Returns the xapan document that was added to the
+        database.
+        """
+        assert self._writable, "Database is not writable"
+        validate = kw.pop('validate', True)
+        schema_type = kw.pop('schema_type', None)
+        if not isinstance(obj, xapian.Document):
+            doc = self.to_document(obj, schema_type=schema_type,
+                                   validate=validate)
+        else:
+            doc = obj
+        self.backend.replace_document(int(docid), doc)
+        return doc
+
+    def to_schema(self, obj, validate=True, schema_type=None):
         """
         Turn an object into an schema instance which is fully
         populated with data from the object.  Optionally validate.
         """
-        if hasattr(obj, '__xodb_schema__'):
-            schema_type = obj.__xodb_schema__
-        else:
-            schema_type = self.schema_for(type(obj))
+        if not schema_type:
+            if hasattr(obj, '__xodb_schema__'):
+                schema_type = obj.__xodb_schema__
+            else:
+                schema_type = self.schema_for(type(obj))
         schema = schema_type.from_defaults()
         schema.update_by_object(obj)
 
@@ -404,12 +464,12 @@ class Database(object):
                                         for c in invalid)))
         return schema
 
-    def to_document(self, obj, validate=True):
+    def to_document(self, obj, validate=True, schema_type=None):
         """
         Convienient wrapper that does the object->schema->document
         transformation.
         """
-        schema = self.to_schema(obj, validate)
+        schema = self.to_schema(obj, validate, schema_type)
         return self.doc_from_dict(schema.memo.dict)
 
     def doc_from_dict(self, data):
@@ -586,7 +646,19 @@ class Database(object):
             # Then string and date
             qp.add_valuerangeprocessor(MultipleValueRangeProcessor(
                 dict(((k, self.values[k])
-                      for k, v in self.value_sorts.items() if v != 'integer')),
+                      for k, v in self.value_sorts.items() if v in
+                        ('string', 'date'))),
+            ))
+            # Serialize date range queries so that they are inclusive.
+            # This allows datetime value range queries to be treated
+            # as [begin,end] rather than [begin,end) as is the default
+            # without these serializers when then datetime range arguments
+            # are not fully qualified.
+            qp.add_valuerangeprocessor(MultipleValueRangeProcessor(
+                dict(((k, self.values[k])
+                      for k, v in self.value_sorts.items() if v == 'datetime')),
+                serializer = lambda x: x + '0'*(14-len(x)),
+                end_serializer = lambda x: x + '9'*(14-len(x))
             ))
         if language in snowball.stoppers:
             qp.set_stemmer(xapian.Stem(language))
@@ -698,7 +770,12 @@ class Database(object):
                         data = self.retry_if_modified(op, retry_limit)
                         typ, data = loads(data)
                         seen.add(docid)
-                        yield self.record_factory(_lookup_schema(typ).from_flat(data))
+                        yield self.record_factory(_lookup_schema(typ).from_flat(data),
+                                                  doc,
+                                                  record.percent,
+                                                  record.rank,
+                                                  record.weight,
+                                                  )
                 # no errors exhuasting the set? break out and we're done
                 break
             except xapian.DatabaseError, e:
@@ -767,7 +844,9 @@ class Database(object):
                                    echo=echo,
                                    retry_limit=retry_limit)
         for facet in suggestions:
-            q = self.querify([query, facet], retry_limit=retry_limit)
+            q = self.querify([query, facet],
+                             retry_limit=retry_limit,
+                             language=language)
             if echo:
                 print str(q)
             else:
@@ -928,14 +1007,24 @@ class Database(object):
         if limit == 0:
             limit = int(self.backend.get_doccount() * klimit)
 
-        eset = enq.get_eset(limit, rset, xapian.Enquire.INCLUDE_QUERY_TERMS,
-                            1.0, decider)
+        if xapian_version <= (1, 2, 4):
+            eset = enq.get_eset(limit, rset,
+                                enq.INCLUDE_QUERY_TERMS,
+                                1.0, decider)
+        elif xapian_version == (1, 2, 5):
+            eset = enq.get_eset(limit, rset,
+                                enq.INCLUDE_QUERY_TERMS,
+                                1.0, decider, -3)
+        else:
+            eset = enq.get_eset(limit, rset,
+                                enq.INCLUDE_QUERY_TERMS,
+                                1.0, decider, 0)
 
         for item in eset.items:
             val = item[0].decode('utf8')
             if prefix and val.startswith('X%s:' % prefix.upper()):
                 suffix = val[len(prefix) + 2:]
-                if ' ' in suffix:
+                if ' ' in suffix or '..' in suffix:
                     val = '%s:"%s"' % (prefix, suffix)
                 else:
                     val = '%s:%s' % (prefix, suffix)
